package ers.nano.Sugma;

/**
 * Author: ER Samson
 * 
 * Iterations built from (ISuckNano 0.1d)
 * 
 * Movement:
 * 		• Mark Whiteley      - http://mark.random-article.com/robocode/improved_movement.html
 * 		• Wiki               - https://robowiki.net/wiki/User:Exauge/snippets
 * 		• Wiki               - https://robowiki.net/wiki/Oscillator_Movement
 * 		• Sheldor/Skilgannon - https://robowiki.net/wiki/Yatagan
 * 		• John Cleland       - https://robowiki.net/wiki/NeophytePRAL
 * 
 * Radar:
 * 		• Wiki               - https://robowiki.net/wiki/One_on_One_Radar
 * 
 * Gun Targetting:
 * 		• Wiki               - https://robowiki.net/wiki/Rolling_Averages
 * 		• Old Wiki           - https://old.robowiki.net/cgi-bin/robowiki?VisitCountStats/LimitFileSize
 *		• Exauge             - https://robowiki.net/wiki/User:Exauge/snippets
 * 		• Sheldor/Skilgannon - https://robowiki.net/wiki/Yatagan
 *      • Simonton           - https://robowiki.net/wiki/WeekendObsession
 * 		• John Cleland       - https://robowiki.net/wiki/NeophytePRAL
 * 
 * Code Shrinking Tips:
 * 		• Sheldor/Skilgannon - https://robowiki.net/wiki/Yatagan
 *      • Simonton           - https://robowiki.net/wiki/WeekendObsession
 * 
 * Symbolic Pattern String (old PM gun):
 *      • Sheldor            - https://robowiki.net/wiki/Foilist
 */

import robocode.*;
import robocode.util.Utils;
//import java.awt.*;

public class SugmaVCS extends AdvancedRobot {

	/* ---------- Gun Variables ---------- */
	private static final int	BULLET_POWER	= 2;
	private static final int	BULLET_VELOCITY	= (int)(20 - 3 * BULLET_POWER);
	private static final int	ANTIRAM_FACTOR	= 100;
	private static final int	LATVEL_BINS		= 17;
	private static final int	LATVEL_OFFSET	= 8;
	private static final int	LATVEL_MAX		= 9;
	private static int[]		latVelBins		= new int[LATVEL_BINS];
	//private static int			highIndex		= 0;

	/* ---------- Movement Variables ---------- */
	//private static final int	ORBIT_DISTANCE	= 180;
	//private static final int	DISTANCE_FACTOR	= 3000;
	private static final double	RAND_DIR_FACTOR	= 0.63;
	private static double 		moveDirection 	= 1;
	//private static double 		moveMode 		= 1;
	//private static double 		enemyHealth;
	//static double myTest;

	public void run() {
		// Color Identification
		//setColors(Color.blue, Color.red, Color.orange);

		// Easy targetting and infinite radar movement
        setAdjustGunForRobotTurn(true); // Saving code size!
		setTurnRadarRightRadians(Double.POSITIVE_INFINITY);
	}

	public void onScannedRobot(ScannedRobotEvent e) {

		/* ---------- Generic Variables ---------- */
		int i = 0;
		double absoluteBearing;
		double angleOffset = 0;
		int index;
		//double enemyDistance;
	
		/* ---------- Musashi Movement Logic ----------	*/
		// Turn perpendicular to enemy, maintain distance, no wall avoidance D:				
		//setTurnRightRadians(Math.cos((absoluteBearing = e.getBearingRadians()) + 
		//		((ORBIT_DISTANCE - (enemyDistance = e.getDistance())) * 
		//		(getVelocity() / DISTANCE_FACTOR)))); // Idea from Yatagan
		setTurnRightRadians(Math.cos((absoluteBearing = e.getBearingRadians()) - 
				(0.025 * getVelocity()))); // Saving code size!	

		/* ---------- Gun Logic ---------- */
		if (latVelBins[index = (int)(e.getVelocity()) + LATVEL_OFFSET] < LATVEL_MAX) {
            latVelBins[index]++;
        } else {	
			do {	
				if (i != index && latVelBins[i] > 0) latVelBins[i]--;				
				angleOffset += latVelBins[i] * (i - LATVEL_OFFSET);
				i++;
			} while (i < latVelBins.length);
		}

		// Aim. Ignore bullet power/speed changes from own gun	
		absoluteBearing += getHeadingRadians();

		setTurnGunRightRadians(Utils.normalRelativeAngle(absoluteBearing - 
				getGunHeadingRadians() + ((angleOffset / LATVEL_MAX  * Math.sin(e.getHeadingRadians() - 
				absoluteBearing)) / BULLET_VELOCITY)));	
		
		// Firing with Anti-Ram and Movement
		//setFire(Math.min(BULLET_POWER + (int)(ANTIRAM_FACTOR / enemyDistance), 
		//		enemyHealth / 4));	
		if (setFireBullet(BULLET_POWER + (int)(ANTIRAM_FACTOR / e.getDistance())) != null) {
			// Direction randomizer, Antiram
			if (Math.random() > RAND_DIR_FACTOR) {			
				onHitWall(null); // Code saving mechanism, idea from Yatagan
			}

			setAhead((Math.random() * 140) * moveDirection);
			//setAhead((Math.random() * 70 + 50) * moveDirection);
			//setAhead(moveDirection *= moveMode);
		}
		//setFire(BULLET_POWER);		
		
		// DEBUGGING
		//out.println(myTest/20);		
        //for (int j = 0; j < latVelBins.length; j++) {
		//	out.println(j - LATVEL_OFFSET + " " + latVelBins[j]);
        //}

		/* ---------- Multi-Mode Movement Logic ---------- */
		// I change strat!
		//if ((char)((enemyHealth - 1.0999 - (enemyHealth = e.getEnergy()))) < 2) {
		//if (enemyHealth > (enemyHealth = e.getEnergy())) {
		//	setAhead(moveDirection *= moveMode);

			//if (matchLength % 3 == 0) { // Change direction randomizer based from PM gun
			//	onHitWall(null); // Code saving mechanism, idea from Yatagan
			//}
		//}

		/* ---------- Radar Logic ---------- */
		// From RoboWiki
		setTurnRadarLeft(getRadarTurnRemaining());		
	}
	
	/**
	 * Reverse direction when I hit wall
	 */
	public void onHitWall(HitWallEvent e) {
		moveDirection = -moveDirection;
	}

	/**
	 * Change movement mode when enemy bullet hits
	 */	
	//public void onHitByBullet(HitByBulletEvent e) {
	//public void onDeath(DeathEvent e) {
	//	moveMode = -moveMode;
	//}
}

