package ers.nano.Sunderer;

/**
 * Author: ER Samson
 * 
 * Iterations built from ISuckNano, Meroka, LigMA, SugmaVCS
 * 
 * Movement:
 * 		• Mark Whiteley      - http://mark.random-article.com/robocode/improved_movement.html
 * 		• Wiki               - https://robowiki.net/wiki/User:Exauge/snippets
 * 		• Wiki               - https://robowiki.net/wiki/Oscillator_Movement
 * 		• Sheldor/Skilgannon - https://robowiki.net/wiki/Yatagan
 * 		• John Cleland       - https://robowiki.net/wiki/NeophytePRAL
 * 
 * Radar:
 * 		• Wiki               - https://robowiki.net/wiki/One_on_One_Radar
 * 
 * Gun Targetting:
 * 		• Wiki               - https://robowiki.net/wiki/Rolling_Averages
 * 		• Old Wiki           - https://old.robowiki.net/cgi-bin/robowiki?VisitCountStats/LimitFileSize
 *		• Exauge             - https://robowiki.net/wiki/User:Exauge/snippets
 * 		• Sheldor/Skilgannon - https://robowiki.net/wiki/Yatagan
 *      • Simonton           - https://robowiki.net/wiki/WeekendObsession
 * 		• John Cleland       - https://robowiki.net/wiki/NeophytePRAL
 * 
 * Code Shrinking Tips:
 * 		• Sheldor/Skilgannon - https://robowiki.net/wiki/Yatagan
 *      • Simonton           - https://robowiki.net/wiki/WeekendObsession
 * 
 * Symbolic Pattern String (old PM gun):
 *      • Sheldor            - https://robowiki.net/wiki/Foilist
 */

import robocode.*;
import robocode.util.Utils;
//import java.awt.*;

public class Sunderer extends AdvancedRobot {
	
	/* ---------- Gun Variables ---------- */
	private static final int	BULLET_POWER	= 2;
	private static final int	BULLET_VELOCITY	= (int)(20 - 3 * BULLET_POWER);
	private static final int	ANTIRAM_FACTOR	= 100;
	private static final int	AVERAGE_FACTOR	= 7;
	private static final int	MAX_BINS		= 10;
	private static double[]		moveBins		= new double[MAX_BINS];
	private static int			moveTimes;
	private static int			index;

	/* ---------- Movement Variables ---------- */
	private static final int	ORBIT_DISTANCE	= 180;
	private static final int	DISTANCE_FACTOR	= 3000;
	private static final double	RAND_DIR_FACTOR	= 0.60;
	private static int 			moveDirection 	= 1;
//	private static int			moveMode		= 1;
	private static double 		enemyHealth;

	public void run() {
		// Color Identification
		//setColors(Color.blue, Color.red, Color.orange);

		// Easy targetting and infinite radar movement
        setAdjustGunForRobotTurn(true); // Saving code size!
		setTurnRadarRightRadians(Double.POSITIVE_INFINITY);
	}

	/**
	 * This is where the magic happens!
	 */
	public void onScannedRobot(ScannedRobotEvent e) {

		/* ---------- Generic Variables ---------- */	
		double 	absoluteBearing;
		double 	enemyDistance;
		double	enemyVelocity;

		/* ---------- Musashi Movement Logic ----------	*/
		// Turn perpendicular to enemy, maintain distance, no wall avoidance D:		
		//setTurnRightRadians(Math.cos((absoluteBearing = e.getBearingRadians()) - (0.2 * moveDirection))); // Saving code size!			
		setTurnRightRadians(Math.cos((absoluteBearing = e.getBearingRadians()) + 
				((ORBIT_DISTANCE - (enemyDistance = e.getDistance())) * 
				(getVelocity() / DISTANCE_FACTOR)))); // Idea from Yatagan

		/* ---------- Gun Logic ---------- */
		// Moving average of lateral velocities of opponent
		// Segmentation - Number of ticks since the velocity was last 0
		absoluteBearing += getHeadingRadians();

		moveBins[index] = (moveBins[index] * AVERAGE_FACTOR + 
				(enemyVelocity = e.getVelocity()) * Math.sin(e.getHeadingRadians() - 
				absoluteBearing)) / (AVERAGE_FACTOR + 1);	

		// I hate occilators!
		moveTimes++;
		if(enemyVelocity == 0) {
			moveTimes = 0;
		} 
		// Index limiter
		index = Math.min(MAX_BINS - 1, moveTimes / (MAX_BINS / 2));

		// Aim. Ignore bullet power/speed changes from own gun. Linear.
		setTurnGunRightRadians(Utils.normalRelativeAngle(absoluteBearing - 
				getGunHeadingRadians() + (moveBins[index] / BULLET_VELOCITY)));

		// Energy management. Anti-Ram.
		//setFire(Math.min(BULLET_POWER + (int)(ANTIRAM_FACTOR / enemyDistance), 
		//		enemyHealth / 4));
		setFire(BULLET_POWER + (int)(ANTIRAM_FACTOR / enemyDistance));

		/* --------- Movement Logic ---------- */
		// Stop'n'Go based on bullet drop
		if (enemyHealth > (enemyHealth = e.getEnergy())) {

			// Direction randomizer
			if (Math.random() > RAND_DIR_FACTOR) { 		
				onHitWall(null); // Code saving mechanism, idea from Yatagan
			}

			// Multi-mode Movement
			//setAhead((moveDirection *= moveMode));
			setAhead(((Math.random() * 70) + 50) * moveDirection);
		}

		/* ---------- Radar Logic ---------- */
		// From RoboWiki
		setTurnRadarLeft(getRadarTurnRemaining());		
	}
	
	/**
	 * Reverse direction when I hit wall
	 */
	public void onHitWall(HitWallEvent e) {
		moveDirection = -moveDirection;
	}

	/**
	 * Change movement mode when enemy bullet hits
	 */	
//	public void onHitByBullet(HitByBulletEvent e) {
	//public void onDeath(DeathEvent e) {
	//	moveMode = -moveMode;
//	}
}


