package kc.mega;
import java.awt.Color;
import java.awt.Graphics2D;

import kc.mega.aim.Aimer;
import kc.mega.game.BattleField;
import kc.mega.game.GameState;
import kc.mega.misc.DatasetWriter;
import kc.mega.misc.Scanner;
import kc.mega.misc.Shielder;
import kc.mega.misc.Strategy;
import kc.mega.move.Mover;
import kc.mega.utils.Painter;
import kc.mega.wave.WaveManager;
import robocode.AdvancedRobot;
import robocode.Bullet;
import robocode.BulletHitBulletEvent;
import robocode.BulletHitEvent;
import robocode.DeathEvent;
import robocode.Event;
import robocode.HitByBulletEvent;
import robocode.HitRobotEvent;
import robocode.ScannedRobotEvent;
import robocode.SkippedTurnEvent;
import robocode.WinEvent;
import wiki.mc2k7.RaikoGun;

/**
 * BeepBoop - a robot by Kevin Clark (Kev)
 * See http://robowiki.net/wiki/BeepBoop for details.
 */
public class BeepBoop extends AdvancedRobot {
  public static boolean TC = false;
  public static boolean MC = false;
  public static boolean MC2k7 = false;
  public static boolean paint = false;
  public static boolean writeData = false;
  public static boolean verbose = false;

  private static GameState gs;
  private static WaveManager waveManager;
  private static Strategy strategy;
  private static Scanner scanner;
  private static Shielder shielder;
  private static Aimer aimer;
  private static Mover mover;
  private static RaikoGun raikoGun;
  public static DatasetWriter datasetWriter;

  private static int[] finishes;
  private static int skippedTurns = 0;

  private ScannedRobotEvent lastScanEvent;
  private Graphics2D lastGraphics;
  private long lastPaintTime;
  private boolean hasWon;

  @Override
  public void run() {
    System.out.println("Beep Boop!");
    setColors(new Color(20, 25, 35), new Color(20, 25, 35), new Color(0, 255, 128));
    setAdjustRadarForGunTurn(true);
    setAdjustGunForRobotTurn(true);

    if(getRoundNum() == 0) {
      gs = new GameState(this);
      scanner = new Scanner(this);
      waveManager = new WaveManager();
      strategy = new Strategy(this, waveManager);
      shielder = new Shielder(this, waveManager.enemyWaveManager);
      aimer = new Aimer(this, strategy, waveManager.myWaveManager, gs.myHitRateTracker);
      mover = new Mover(this, strategy, waveManager.enemyWaveManager, gs.enemyHitRateTracker);
      datasetWriter = new DatasetWriter();
      finishes = new int[getOthers() + 1];
      if (MC2k7) {
        TC = false;
        MC = true;
        raikoGun = new RaikoGun(this);
      }
      if (TC || MC) {
        strategy.shield = false;
      }
    }

    BattleField.INSTANCE.init(this);
    gs.init();
    waveManager.init();
    scanner.init();
    aimer.init();
    mover.init();
    if (strategy.shield) {
      shielder.init();
    }

    do {
      if ((getOthers() == 0 || scanner.getLastScannedTime() < getTime()) && lastScanEvent != null) {
        gs.updateNoScan();
        strategy.update(gs);
        if (!TC && !strategy.shield) {
          mover.move(gs);
        }
      }
      scanner.sweep(gs);
      paint = lastPaintTime == getTime();
      if (paint && lastGraphics != null) {
        Painter.THIS_TICK.paint(lastGraphics);
        Painter.PERMANENT.paint(lastGraphics);
      }
      Painter.THIS_TICK.clear();
      execute();
    } while(true);
  }

  @Override
  public void onScannedRobot(ScannedRobotEvent e) {
    if (writeData) {
      datasetWriter.init(this, e.getName());
    }
    lastScanEvent = e;
    gs.update(e);
    strategy.update(gs);
    scanner.scan(gs);
    if (strategy.shield) {
      shielder.shield(gs);
    } else {
      if (!TC) {
        mover.move(gs);
      }
      if (!MC) {
        aimer.aim(gs);
      }
      if (MC2k7) {
        raikoGun.onScannedRobot(e);
      }
    }
  }

  @Override
  public void onBulletHit(BulletHitEvent e) {
    if (!hasWon) {
      if (!MC && !strategy.shield) {
        aimer.onBulletHit(e);
      }
      gs.onBulletHit(e);
    }
  }

  @Override
  public void onHitByBullet(HitByBulletEvent e) {
    if (strategy.shield) {
      shielder.onHitByBullet(e, gs);
    } else if (!TC) {
      mover.onHitByBullet(e, gs);
    }
    gs.onHitByBullet(e);
  }

  @Override
  public void onBulletHitBullet(BulletHitBulletEvent e) {
    if (strategy.shield) {
      shielder.onBulletHitBullet(e);
    } else {
      if (!TC) {
        mover.onBulletHitBullet(e);
      }
      if (!MC) {
        aimer.onBulletHitBullet(e);
      }
      waveManager.onBulletHitBullet();
    }
  }

  @Override
  public void onHitRobot(HitRobotEvent e) {
    gs.onHitRobot(e);
  }

  @Override
  public void onSkippedTurn(SkippedTurnEvent e) {
    skippedTurns++;
    if (gs.gameTime != getTime()) {
      gs.update(lastScanEvent);
    }
  }

  @Override
  public void onWin(WinEvent e) {
    for (Event otherEvent: getAllEvents()) {
      if(otherEvent instanceof BulletHitEvent) {
        onBulletHit((BulletHitEvent)otherEvent);
      }
    }
    hasWon = true;
    onRoundEnd();
  }

  @Override
  public void onDeath(DeathEvent e) {
    if (!hasWon) {
      onRoundEnd();
      System.out.println();
    }
  }

  public void onRoundEnd() {
    finishes[getOthers()]++;
    if (writeData) {
      datasetWriter.onRoundEnd();
    }
    if (!strategy.shield) {
      if (!MC) {
        aimer.printStats();
      }
      if (!TC) {
        mover.printStats();
      }
    }
    System.out.println("Skipped Turns: " + skippedTurns);
    System.out.print("Finishes: ");
    for (int i = 0; i < finishes.length; i++) {
      System.out.print((i == getOthers() ? "*" : "") + finishes[i] + " ");
    }
    System.out.println();
  }

  @Override
  public void onPaint(Graphics2D g) {
    lastPaintTime = getTime();
    this.lastGraphics = g;
  }

  @Override
  public void setFire(double power) {
    Bullet b = setFireBullet(power);
    if (b != null && MC2k7) {
      gs.onMyFire(power);
    }
  }
}
