package kc.mega.aim;

import kc.mega.BeepBoop;
import kc.mega.game.BotState;
import kc.mega.game.GameState;
import kc.mega.game.HitRateTracker;
import kc.mega.learning.AimModel;
import kc.mega.misc.Strategy;
import kc.mega.utils.Geom;
import kc.mega.wave.WaveManager;
import kc.mega.wave.WaveWithFeatures;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import robocode.AdvancedRobot;
import robocode.BulletHitBulletEvent;
import robocode.BulletHitEvent;
import robocode.util.Utils;

public class Aimer {
  private final AdvancedRobot bot;
  private final Strategy strategy;
  private final WaveManager.MyWaveManager waveManager;
  private final HitRateTracker hitRateTracker;
  private final List<AimModel> aimModels;

  private boolean hasFired;
  private AimModel firingModel;
  private List<Double> lastAimAngles;

  public Aimer(AdvancedRobot bot, Strategy strategy, WaveManager.MyWaveManager waveManager,
      HitRateTracker hitRateTracker) {
    this.bot = bot;
    this.strategy = strategy;
    this.waveManager = waveManager;
    this.hitRateTracker = hitRateTracker;
    aimModels = Arrays.asList(AimModels.getMainAimModel(), AimModels.getAntiSurferAimModel());
    firingModel = aimModels.get(0);
  }

  public void init() {
    hasFired = false;
    if (!strategy.shield) {
      System.out.println("Using " + firingModel.name + " gun");
    }
  }

  public void aim(GameState gs) {
    BotState myState = gs.getMyState();
    BotState enemyState = gs.getEnemyState();
    for (AimWave w : waveManager.updateWaves(enemyState)) {
      for (AimModel aimModel : aimModels) {
        aimModel.train(w);
      }
      if (w.hasBullet || (w.virtuality == 0 && w.didCollide)) {
        hitRateTracker.logShotPassed(w.power);
      }
      /*if (BeepBoop.writeData) {
        datasetWriter.write("gun-data", w.toArray());
      }*/
    }

    double bulletPower = strategy.ram ? 0 : Math.min(firingModel.hasData() ?
        3 : 0.15, BulletPowerSelector.bestBulletPower(gs, strategy));
    boolean doFire = false;
    if (bulletPower > 0.099999 &&
        myState.gunHeat <= 0.0 &&
        Math.abs(bot.getGunTurnRemainingRadians()) < 0.0001) {
      hasFired = true;
      doFire = true;
      gs.onMyFire(bulletPower);
      gs.bot.setFireBullet(bulletPower);
    }

    if (gs.myHistory.size() <= 2) {
      return;
    }

    Point2D.Double enemyLocation = enemyState.location;
    bot.setTurnGunRightRadians(Utils.normalRelativeAngle(Geom.absoluteBearing(
        gs.myFuture.get(0).location, enemyLocation) - bot.getGunHeadingRadians()));

    int ticksUntilFire = (int)Math.ceil(myState.gunHeat / bot.getGunCoolingRate());
    if (!doFire && ticksUntilFire <= 3) {
      List<BotState> myNextHistory = new ArrayList<>(gs.myHistory);
      List<BotState> enemyNextHistory = new ArrayList<>(gs.enemyHistory);
      myNextHistory.add(0, gs.myFuture.get(0));
      enemyNextHistory.add(0, gs.enemyFuture.get(0));
      WaveWithFeatures aimWave = (WaveWithFeatures)new WaveWithFeatures.Builder()
          .waveManager(waveManager)
          .myHistory(myNextHistory)
          .enemyHistory(enemyNextHistory)
          .power(bulletPower)
          .hasBullet(true)
          .build();
      double aimAngle = ticksUntilFire > 1 ? firingModel.getAimAngleFast(aimWave) :
        firingModel.getAimAngle(aimWave);
      bot.setTurnGunRightRadians(Utils.normalRelativeAngle(aimAngle - bot.getGunHeadingRadians()));
    } else {
      pickAimModel();
    }

    if (hasFired && bulletPower > 0.099999) {
      AimWave w = (AimWave)new AimWave.Builder()
          .gunHeading(bot.getGunHeadingRadians())
          .aimAngles(lastAimAngles)
          .waveManager(waveManager)
          .myHistory(gs.myHistory)
          .enemyHistory(gs.enemyHistory)
          .power(bulletPower)
          .hasBullet(doFire)
          .build();
      waveManager.add(w);
    }
  }

  public void pickAimModel() {
    int bestModelInd = 0;
    if (hitRateTracker.hitRateInBounds(0, 0.12)) {
      bestModelInd = 1;
    }
    if (aimModels.get(bestModelInd) != firingModel) {
      firingModel = aimModels.get(bestModelInd);
      if (BeepBoop.verbose) {
        System.out.println("Switching to " + firingModel.name + " gun");
      }
    }
  }

  public void onBulletHit(BulletHitEvent e) {
    AimWave hitWave = waveManager.getHitWave(e.getBullet().getPower());
    if (hitWave != null) {
      for (AimWave w : waveManager.getWaves()) {
        if (!w.hasBullet && Math.abs(w.fireTime - hitWave.fireTime) <= 3) {
          w.didHit = true;
        }
      }
    }
  }

  public void onBulletHitBullet(BulletHitBulletEvent e) {
    if (bot.getOthers() > 0) {
      AimWave hitWave = waveManager.getBulletHitBulletWave(
          new Point2D.Double(e.getBullet().getX(), e.getBullet().getY()),
          e.getBullet().getPower(), bot.getTime());
      if (hitWave != null) {
        for (AimWave w : waveManager.getWaves()) {
          if (!w.hasBullet && Math.abs(w.fireTime - hitWave.fireTime) <= 3) {
            w.didCollide = true;
          }
        }
      }
    }
  }

  public void printStats() {
    System.out.println("My Hit Rate: " + hitRateTracker.getHitRateStr());
    if (BeepBoop.TC) {
      System.out.println("TC Score: " + (float)(
          hitRateTracker.getDamage() / (1 + bot.getRoundNum())));
    }
  }
}
