package kc.mega.game;

import kc.mega.utils.KUtils;

import java.awt.geom.Point2D;


public class BotState extends PredictState {
  public long lastFireTime;
  public int shotsFired;
  public double gunHeat;
  public double energy;

  public BotState(Point2D.Double location, double heading, double velocity, double energy,
                   double gunHeat, int shotsFired, long lastFireTime, long gameTime) {
    super(location, heading, velocity, gameTime);
    this.energy = energy;
    this.gunHeat = gunHeat;
    this.shotsFired = shotsFired;
    this.lastFireTime = lastFireTime;
  }

  @Override
  public BotState getNextState(int direction, double turn) {
    return getNextState(super.getNextState(direction, turn));
  }

  @Override
  public BotState getNextState(double turn) {
    return getNextState(super.getNextState(turn));
  }

  @Override
  public BotState predictNextState(PredictState previous) {
    return getNextState(super.predictNextState(previous));
  }

  private BotState getNextState(PredictState predicted) {
    return new BotState(
        predicted.location, predicted.heading, predicted.velocity,
        energy, gunHeat - BattleField.INSTANCE.getGunCoolingRate(),
        shotsFired, lastFireTime, gameTime + 1);
  }

  public PredictState asPredictState() {
    return new PredictState(location, heading, velocity, gameTime);
  }

  public void updateFromFire(double power) {
    gunHeat = KUtils.gunHeat(power);
    energy -= power;
    lastFireTime = gameTime;
    shotsFired += 1;
  }

  public void updateFromBulletHit(double power) {
    energy += power * 3;
  }

  public void updateFromHitByBullet(double power) {
    energy -= KUtils.bulletDamage(power);
  }
}
