/*
Sabreuse v1.0.0 by Sheldor.  05/27/2013
A NanoBot with stop and go/ramming movement + averaged linear targeting.
Codesize: 246 Bytes without any colors.

Sabre is one of the three forms of modern sport fencing,
along with Epee and Foil.  http://en.wikipedia.org/wiki/Sabre_%28fencing%29

Credits:
Thanks go to the authors of the following bots: 
jk.nano.Machete, nz.jdc.NeophyteSRAL, jk.sheldor.nano.Yatagan.
Also, a general thanks to all open source bot authors and contributors to the RoboWiki.

Sabreuse is open source and released under the terms of the RoboWiki Public Code License (RWPCL) - Version 1.1.
see license here:  http://robowiki.net/wiki/RWPCL
*/

package sheldor.nano;

import robocode.*;
import robocode.util.Utils;

public class Sabreuse extends AdvancedRobot
{
	//Global variables.
	static double direction = 1;
	static double enemyEnergy;
	static double averageVelocity;
	static double rollingDepth = 30;
		
	//En garde!
	public void run()
	{
		//Start spinning radar.
		setTurnRadarRightRadians(Double.POSITIVE_INFINITY);
	}

	public void onScannedRobot(ScannedRobotEvent e)
	{
		//Local variables.
		double absoluteBearing;
		double bearing;
		double lead;
				
		//In stop and go mode, keep a rolling average of the last 30 enemy lateral velocities.
		//In ramming mode, keep a rolling average of the last 0 enemy lateral velocities.
		//The latter has the same effect as regular linear targeting.
		setTurnGunRightRadians(Utils.normalRelativeAngle((absoluteBearing = getHeadingRadians() + (bearing = e.getBearingRadians())) +
			(lead = Math.asin((averageVelocity = ((averageVelocity * rollingDepth) + (e.getVelocity() * Math.sin(e.getHeadingRadians() - absoluteBearing))) / (1 + rollingDepth))
			/ 14)) - getGunHeadingRadians()));
						
		//Stop and Go movement with variable length.
		//The absolute bearing variable is re-used here to represent the change in the enemy's energy.  This saves two bytes.
		if ((absoluteBearing = (enemyEnergy - (enemyEnergy = e.getEnergy()))) > 0)
		{
			setAhead(direction * (3 + ((int)(absoluteBearing / 0.5000001)) << 3));
		}
		
		//Stay mostly perpendicular to the enemy, but try to maintain a distance of 225 pixels.
		//Here the absolute bearing variable is re-used yet again to represent the enemy distance.  This saves one byte.
		setTurnRightRadians(Math.cos(bearing - ((absoluteBearing = e.getDistance()) - 225) * direction / 650));
						
		//Fire medium power bullets most of the time, but use full power at very close range.
		setFire(2 + (100 / (int)absoluteBearing));
		
		//If the bot loses in the first four rounds, switch from stop and go to ramming movement.
		//Notice that I use the same linear lead for ramming as I use in the gun code.
		if (rollingDepth == 0)
		{
			setTurnRightRadians(Math.tan(bearing += lead));
			setAhead(Math.cos(bearing) / 0);
		}
	
		//Infinite radar lock.
		setTurnRadarLeftRadians(getRadarTurnRemainingRadians());
	}

	public void onBulletHit (BulletHitEvent e)
	{		
		//Adjust the enemy energy variable when the bot hits the enemy.
		//This makes a big difference against linear targeting.
		enemyEnergy -= 10;
	}

	public void onDeath(DeathEvent e)
	{
		//If the stop and go movement loses in the first four rounds, switch to ramming movement.
		if (getRoundNum() < 4)
		{
			rollingDepth = 0;
		}
	}
	
	public void onHitWall(HitWallEvent e)
	{
		//Reverse movement direction when the bot hits a wall.
		direction = -direction;
	}
}																																																						